from datetime import date, datetime, time
from typing import Union, Tuple, Dict, Optional

SQLParameter = Optional[Union[str, int, float, date, datetime, time, bool, bytes]]
SQLParameters = Union[Tuple[SQLParameter, ...], Dict[str, SQLParameter], SQLParameter]



import logging
import re
import struct
from datetime import date, time, datetime, timedelta, timezone
from decimal import Decimal
from typing import Dict, Any, List, Union

import pymssql as sql
from dateutil.parser import isoparse

from pymssqlutils.helpers import SQLParameter

logger = logging.getLogger(__name__)

date_regex = re.compile(r"(\d{4})-(\d{2})-(\d{2})")
time_regex = re.compile(r"(\d{2}):(\d{2}):(\d{2})(?:\.(\d{1,6})\d?)?")
datetime2_regex = re.compile(r"\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}.\d{7}")
datetimeoffset_regex = re.compile(
    r"(\d{4}-\d{2}-\d{2})\s(\d{2}:\d{2}:\d{2}(?:\.\d{1,6})?)\d?\s([+-]\d{2}:\d{2})"
)


def _is_time(x):
    return re.fullmatch(time_regex, x)


def _is_date(x):
    return re.fullmatch(date_regex, x)


def _is_datetimeoffset(x):
    return re.fullmatch(datetimeoffset_regex, x)


def _is_datetime2(x):
    return re.fullmatch(datetime2_regex, x)


def _clean(item: Any) -> Any:
    if isinstance(item, str):
        # handle various date/time returns that pymssql doesn't handle
        match = _is_date(item)
        if match:
            return date(int(match[1]), int(match[2]), int(match[3]))

        match = _is_time(item)
        if match:
            return time(
                int(match[1]),
                int(match[2]),
                int(match[3]),
                microsecond=int(match[4]) if match[4] else 0,
            )

        match = _is_datetime2(item)
        if match:
            return isoparse(item[:-1])

        match = _is_datetimeoffset(item)
        if match:
            return isoparse(f"{match[1]}T{match[2]}" + (match[3] or ""))

    if isinstance(item, bytes):
        try:
            # fix for certain versions of FreeTDS driver returning Datetimeoffset as bytes
            microseconds, days, tz, _ = struct.unpack("QIhH", item)
            return datetime(
                1900, 1, 1, 0, 0, 0, tzinfo=timezone(timedelta(minutes=tz))
            ) + timedelta(days=days, minutes=tz, microseconds=microseconds / 10)
        except struct.error:
            return item

    if isinstance(item, Decimal):
        return float(item)

    return item


class DatabaseError(Exception):
    pass


class DatabaseResult:
    ok: bool
    fetch: bool
    commit: bool
    columns: List[str] = None
    data: List[Dict[str, SQLParameter]] = None
    error: sql.Error = None

    def __init__(
        self,
        ok: bool,
        fetch: bool,
        commit: bool,
        data: List[Dict] = None,
        error: sql.Error = None,
    ):
        self.ok = ok
        self.fetch = fetch
        self.commit = commit

        if error:
            self.error = error
        else:
            if data:
                self.columns = [*data[0]]
                self.data = [{k: _clean(v) for k, v in row.items()} for row in data]
            elif self.fetch:
                self.data = []
                self.columns = []

    def write_error_to_logger(self, name: str = "unknown") -> None:
        """
        Writes the error to logger.

        :param name: str, an optional name to show in the error string.
        :return: None
        """
        if self.ok:
            raise ValueError("This execution did not error")
        error_text = str(
            self.error.args[1] if len(self.error.args) >= 2 else self.error
        )
        logger.error(
            f"DatabaseResult Error (<{name}|fetch={self.fetch},commit={self.commit}>)"
            f": <{type(self.error).__name__}> {error_text}"
        )

    def raise_error(self, name: str = "unknown") -> None:
        """
        Raises a pymssql DatabaseError with an optional name to help identify the operation.
        :param name: str, an optional name to show in the error string.
        :return: None
        """
        if self.ok:
            raise ValueError("This execution did not error")
        raise DatabaseError(
            f"<{name}|fetch={self.fetch},commit={self.commit}> bad execution"
        ) from self.error

    def to_dataframe(self, *args, **kwargs):
        """
        Return the data as a Pandas DataFrame, all args and kwargs are passed to
        the DataFrame initiation method.
        :return: a DataFrame
        """
        if self.data is None:
            raise ValueError("DatabaseResult class has no data to cast to DataFrame")

        # noinspection PyUnresolvedReferences
        try:
            from pandas import DataFrame
        except ImportError:
            raise RuntimeError("Pandas must be installed to use this method")

        return DataFrame(data=self.data, *args, **kwargs)

    def to_json(self, as_bytes=False) -> Union[bytes, str]:
        """
        returns the serialized data as a JSON format string.
        :params as_bytes: bool, if True returns the JSON object as UTF-8 encoded bytes instead of string
        :return: Union[bytes, str]
        """

        if self.data is None:
            raise ValueError("DatabaseResult class has no data to cast to DataFrame")

        # noinspection PyUnresolvedReferences
        try:
            from orjson import dumps
        except ImportError:
            raise RuntimeError(
                "ORJSON must be installed to use this method, you can install "
                + "this by running `pip install --upgrade pymssql-utils[json]`"
            )

        if as_bytes:
            return dumps(self.data)
        else:
            return dumps(self.data).decode("UTF-8")



import logging
import os
from itertools import zip_longest
from typing import (
    Dict,
    List,
    Tuple,
    Union,
    Iterable,
    Mapping,
    Any,
)

import pymssql as sql

from .databaseresult import DatabaseResult
from .helpers import SQLParameters, SQLParameter

logger = logging.getLogger(__name__)


def substitute_parameters(operation: str, parameters: SQLParameters) -> str:
    """
    This function returns the SQL operation that would be executed on the server (i.e. after
    parsing and substituting of parameters). Useful for logging & debugging.

    :param operation: The SQL operation requiring substitution
    :param parameters: The parameters to substitute in
    :return: The parameter substituted SQL operation as a string
    """
    if isinstance(parameters, tuple):
        parameters = tuple(
            item.isoformat() if hasattr(item, "isoformat") else item
            for item in parameters
        )
    elif isinstance(parameters, dict):
        parameters = {
            key: item.isoformat() if hasattr(item, "isoformat") else item
            for key, item in parameters.items()
        }
    else:
        if hasattr(parameters, "isoformat"):
            parameters = parameters.isoformat()
        elif parameters is None:
            parameters = (None,)

    return sql._mssql.substitute_params(operation, parameters).decode("UTF-8")


def set_connection_details(
    server: str = None, database: str = None, user: str = None, password: str = None
) -> None:
    """
    Sets the relevant environment variable for each passed parameter
    :param server: the network address of the SQL server to connect to, sets 'MSSQL_SERVER' in the environment
    :param database: the default database to use on the SQL server, sets 'MSSQL_DATABASE' in the environment
    :param user: the user to authenticate against the SQL server with, sets 'MSSQL_USER' in the environment
    :param password: the password to authenticate against the SQL server with, sets 'MSSQL_PASSWORD' in the environment
    :return:
    """
    if server:
        os.environ["MSSQL_SERVER"] = server
    if database:
        os.environ["MSSQL_DATABASE"] = database
    if user:
        os.environ["MSSQL_USER"] = user
    if password:
        os.environ["MSSQL_PASSWORD"] = password


def _with_conn_details(kwargs: Dict) -> Dict:
    if not kwargs:
        kwargs = {}

    kwargs["server"] = kwargs.get("server", os.environ.get("MSSQL_SERVER"))
    kwargs["database"] = kwargs.get("database", os.environ.get("MSSQL_DATABASE"))
    kwargs["user"] = kwargs.get("user", os.environ.get("MSSQL_USER"))
    kwargs["password"] = kwargs.get("password", os.environ.get("MSSQL_PASSWORD"))

    if not kwargs["server"]:
        raise ValueError(
            "server must be passed as a parameter or MSSQL_SERVER must be set in the environment"
        )

    return kwargs


def query(
    operation: str,
    parameters: SQLParameters = None,
    raise_errors: bool = True,
    **kwargs,
) -> DatabaseResult:
    """
    Execute a SQL Operation which DOES NOT COMMIT the transaction & returns the result.

    **kwargs are passed through to the pymssql.connect() method.

    :param operation: the SQL Operation to execute
    :type operation: str
    :param parameters: parameters to substitute into the operation.
    :type parameters: SQLParameters
    :param raise_errors: if True raises errors, else DatabaseResult class will contain the error details
    :type raise_errors: bool, optional
    :return: a DatabaseResult class.
    :rtype: DatabaseResult
    """
    try:
        return _execute(
            [operation],
            [parameters] if parameters else None,
            fetch=True,
            **_with_conn_details(kwargs),
        )
    except sql.Error as err:
        if raise_errors:
            raise err
        return DatabaseResult(ok=False, fetch=True, commit=False, error=err)


def execute(
    operations: Union[str, List[str]],
    parameters: Union[SQLParameters, List[SQLParameters]] = None,
    batch_size: int = None,
    fetch: bool = False,
    raise_errors: bool = True,
    **kwargs,
) -> DatabaseResult:
    """
    Used for a SQL Operation/s which COMMIT the transaction. There are two ways of using this function:

    Passing in a single operation (str) to operations:
        * If parameters is singular, this calls pymssql.execute() and executes a single operation
        * If parameters is plural, this calls pymssql.execute_many() and executes one execution per parameter set

    Passing in multiple operations (List[str]) to operations:
        * If parameters is None, this calls pymssql.execute_many() and executes one execution per operation
        * If parameters is the same length as operations, this calls pymssql.execute() multiple times
          and executes one execution per operation.

    Optionally batch_size can be specified to use string concatenation to batch the operations, this can
    provide significant performance gains if executing 100+ small operations. This is similar to fast_executemany
    found in pyodbc package. A value of 500-1000 is a good default.

    Fetch can be set to True, this will only return the last result set.

    :param operations: the SQL Operation/s to execute. If this is a list then parameters needs to be None
        or a list of the same length.
    :type operations: Union[str, List[str]]
    :param parameters: parameters to substitute into the operation. These can be a single value, tuple or dictionary.
        If operations is a list this parameter needs to either be None or a list of the same length.
    :type parameters: Union[SQLParameters, List[SQLParameters]], optional
    :param batch_size: If specified concatenate the operations together according to the batch_size,
        raises an error if set to True and both operations and parameters are singular
    :type batch_size: int, optional
    :param fetch: return the LAST result of the execution
    :type fetch: bool, optional
    :param raise_errors: if True raises errors, else DatabaseResult class will contain the error details
    :type raise_errors: bool, optional
    :return: a DatabaseResult class
    :rtype: DatabaseResult
    """
    if isinstance(operations, str):
        operations = [operations]
    if parameters is not None and not isinstance(parameters, list):
        parameters = [parameters]

    singular_operations = len(operations) == 1
    singular_parameters = parameters and len(parameters) == 1

    # validate
    if batch_size is not None:
        if singular_operations and singular_parameters:
            raise ValueError(
                "batch_size cannot be used if both operations and parameters are singular"
            )
        if batch_size <= 0:
            raise ValueError("batch_size cannot be negative")

    if parameters is not None:
        if not (singular_parameters or singular_operations) and len(operations) != len(
            parameters
        ):
            raise ValueError(
                "parameters must be the same length as operations if they are both lists"
            )

    try:
        if batch_size:
            return _execute_batched(
                operations, parameters, batch_size, fetch, **_with_conn_details(kwargs)
            )
        return _execute(
            operations,
            parameters,
            commit=True,
            fetch=fetch,
            **_with_conn_details(kwargs),
        )
    except sql.Error as err:
        if raise_errors:
            raise err
        return DatabaseResult(
            ok=False,
            fetch=fetch,
            commit=True,
            error=err,
        )


def _execute_batched(
    operations: List[str],
    parameters: List[SQLParameters] = None,
    batch_size: int = 1000,
    fetch: bool = False,
    **kwargs,
) -> DatabaseResult:
    """
    This is an internal method and you should call execute() instead
    """
    if parameters:
        fillvalue = (
            parameters[-1] if len(parameters) < len(operations) else operations[-1]
        )
        batched = [
            "\n;".join(
                substitute_parameters(operation, parameter_set)
                for operation, parameter_set in zip_longest(
                    operations[i : i + batch_size],
                    parameters[i : i + batch_size],
                    fillvalue=fillvalue,
                )
            )
            for i in range(0, max(len(parameters), len(operations)), batch_size)
        ]
    else:
        batched = [
            "\n;".join(operation for operation in operations[i : i + batch_size])
            for i in range(0, len(operations), batch_size)
        ]

    with sql.connect(as_dict=True, **_with_conn_details(kwargs)) as cnxn:
        with cnxn.cursor() as cur:
            for batch in batched:
                cur.execute(batch)
            data = cur.fetchall() if fetch else None
        cnxn.commit()
    return DatabaseResult(data=data, ok=True, fetch=fetch, commit=True)


def _execute(
    operations: List[str],
    parameters: List[SQLParameters] = None,
    commit: bool = False,
    fetch: bool = False,
    **kwargs,
) -> DatabaseResult:
    """
    This is an internal method and you should call execute() instead
    """
    with sql.connect(as_dict=True, **kwargs) as cnxn:
        with cnxn.cursor() as cur:
            if parameters:
                fillvalue = (
                    parameters[-1]
                    if len(parameters) < len(operations)
                    else operations[-1]
                )
                for operation, parameter_set in zip_longest(
                    operations, parameters, fillvalue=fillvalue
                ):
                    cur.execute(substitute_parameters(operation, parameter_set))
            else:
                for operation in operations:
                    cur.execute(operation)

            data = cur.fetchall() if fetch else None

        if commit:
            cnxn.commit()

    return DatabaseResult(data=data, ok=True, fetch=fetch, commit=commit)


def to_sql_list(listlike: Iterable[SQLParameter]) -> str:
    """
    Transforms an iterable to a SQL list string.
    Intended to be used with the SQL 'in' operator when building dynamic SQL queries.

    e.g. [1, 2, 3] -> '(1, 2, 3)'

    :param listlike: The iterable of SQLParameter to transform
    :return: str
    """
    out_str = ", ".join(substitute_parameters("%s", x) for x in listlike)
    return f"({out_str})"


def model_to_values(
    model: Any,
    prepend: List[Tuple[str, str]] = None,
    append: List[Tuple[str, str]] = None,
) -> str:
    """
    Transforms a Dict or Mapping into a string of the form: '([attr1], [attr2], ...) VALUES (val1, val2, ...)'.
    Intended to be used when creating dynamic SQL INSERT statements.

    Prepend and append can be used to add a SQL column with a static or variable value
    at the beginning or end of the values list.

    e.g. passing prepend = [('prependedColumn', '@prependedColumn')] would return a string of the form:
    '([prependedColumn], [attr1], [attr2], ...) VALUES (@prependedColumn, val1, val2, ...)'

    Note: prepended and appended columns are not parameter substituted, this can leave your code open to sql injection.

    :param model: a Dictionary or anything that implements the __dict__ method (e.g. a Pydantic Model)
    :param prepend: prepend a variable number of columns to the beginning of the values statement.
    :param append: append a variable number of columns to the end of the values statement.
    :return: str
    """
    keys = [x[0] for x in prepend] if prepend else []
    prepend_values = [x[1] for x in prepend] if prepend else []
    append_values = [x[1] for x in append] if append else []

    if isinstance(model, dict):
        keys.extend(model.keys())
        values = list(model.values())
    else:
        keys.extend(model.__dict__.keys())
        values = list(model.__dict__.values())

    if append:
        keys.extend(x[0] for x in append)

    values = ", ".join(
        [
            *map(str, prepend_values),
            # keep (x,) due to: https://github.com/pymssql/pymssql/issues/696
            *[substitute_parameters("%s", (x,)) for x in values],
            *map(str, append_values),
        ]
    )

    column_names = "(" + ", ".join(f"[{key}]" for key in keys) + ")"
    return f"{column_names} VALUES ({values})"



from .databaseresult import DatabaseResult, DatabaseError
from .methods import (
    execute,
    query,
    to_sql_list,
    model_to_values,
    substitute_parameters,
    set_connection_details,
)

__all__ = [
    "execute",
    "query",
    "to_sql_list",
    "model_to_values",
    "substitute_parameters",
    "set_connection_details",
    "DatabaseResult",
    "DatabaseError",
]



import pathlib
from setuptools import setup, find_packages

extra_json = ["orjson>=3.4.0,<4.0.0"]
extra_test = [
    "pytest-mock>=3.6.0,<4.0.0",
    "pandas>=1.0.0,<2.0.0",
    "pytest-dotenv>=0.5.0,<1.0.0",
]
extra_all = extra_json + extra_test

setup(
    name="pymssql-utils",
    version="0.0.16",
    description="pymssql-utils is a small library that wraps pymssql to make your life easier.",
    long_description=(pathlib.Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/invokermain/pymssql-utils",
    author="Tim OSullivan",
    author_email="tim@lanster.dev",
    license="GNU LGPLv2.1",
    classifiers=[
        "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Database",
        "Typing :: Typed",
    ],
    packages=find_packages(),
    python_requires=">=3.6",
    include_package_data=True,
    install_requires=[
        "pymssql>=2.1.0,<3.0.0",
        "python-dateutil>=2.8.0,<3.0.0",
    ],
    extras_require={
        "all": extra_all,
        "json": extra_json,
        "test": extra_test,
    },
)
